@@ -1,3 +1,14 @@
+from typing import Optional, IO, Union
+from pathlib import Path
+import vamb.vambtools as _vambtools
+from torch.utils.data.dataset import TensorDataset as _TensorDataset
+from torch.utils.data import DataLoader as _DataLoader
+from torch.nn.functional import softmax as _softmax
+from torch.optim import Adam as _Adam
+from torch import Tensor
+from torch import nn as _nn
+from math import log as _log
+
 __doc__ = """Encode a depths matrix and a tnf matrix to latent representation.
 
 Creates a variational autoencoder in PyTorch and tries to represent the depths
@@ -5,7 +16,7 @@ and tnf in the latent space under gaussian noise.
 
 Usage:
 >>> vae = VAE(nsamples=6)
->>> dataloader, mask = make_dataloader(depths, tnf)
+>>> dataloader, mask = make_dataloader(depths, tnf, lengths)
 >>> vae.trainmodel(dataloader)
 >>> latent = vae.encode(dataloader) # Encode to latent representation
 >>> latent.shape
@@ -16,21 +27,33 @@ __cmd_doc__ = """Encode depths and TNF using a VAE to latent representation"""
 
 import numpy as _np
 import torch as _torch
+
 _torch.manual_seed(0)
 
-from math import log as _log
+def set_batchsize(data_loader: _DataLoader, batch_size: int, encode=False) -> _DataLoader:
+    """Effectively copy the data loader, but with a different batch size.
 
-from torch import nn as _nn
-from torch.optim import Adam as _Adam
-from torch.nn.functional import softmax as _softmax
-from torch.utils.data import DataLoader as _DataLoader
-from torch.utils.data.dataset import TensorDataset as _TensorDataset
-import vamb.vambtools as _vambtools
+    The `encode` option is used to copy the dataloader to use before encoding.
+    This will not drop the last minibatch whose size may not be the requested
+    batch size, and will also not shuffle the data.
+    """
+    return _DataLoader(
+        dataset=data_loader.dataset,
+        batch_size= batch_size,
+        shuffle=not encode,
+        drop_last=not encode,
+        num_workers=1 if encode else data_loader.num_workers,
+        pin_memory=data_loader.pin_memory,
+    )
 
-if _torch.__version__ < '0.4':
-    raise ImportError('PyTorch version must be 0.4 or newer')
-
-def make_dataloader(rpkm, tnf, batchsize=256, destroy=False, cuda=False):
+def make_dataloader(
+    rpkm: _np.ndarray,
+    tnf: _np.ndarray,
+    lengths: _np.ndarray,
+    batchsize: int = 256,
+    destroy: bool = False,
+    cuda: bool = False,
+) -> tuple[_DataLoader[tuple[Tensor, Tensor, Tensor]], _np.ndarray]:
     """Create a DataLoader and a contig mask from RPKM and TNF.
 
     The dataloader is an object feeding minibatches of contigs to the VAE.
@@ -51,64 +74,94 @@ def make_dataloader(rpkm, tnf, batchsize=256, destroy=False, cuda=False):
     """
 
     if not isinstance(rpkm, _np.ndarray) or not isinstance(tnf, _np.ndarray):
-        raise ValueError('TNF and RPKM must be Numpy arrays')
+        raise ValueError("TNF and RPKM must be Numpy arrays")
 
     if batchsize < 1:
-        raise ValueError('Minimum batchsize of 1, not {}'.format(batchsize))
+        raise ValueError(f"Batch size must be minimum 1, not {batchsize}")
 
-    if len(rpkm) != len(tnf):
-        raise ValueError('Lengths of RPKM and TNF must be the same')
+    if len(rpkm) != len(tnf) or len(tnf) != len(lengths):
+        raise ValueError("Lengths of RPKM, TNF and lengths arrays must be the same")
 
     if not (rpkm.dtype == tnf.dtype == _np.float32):
-        raise ValueError('TNF and RPKM must be Numpy arrays of dtype float32')
+        raise ValueError("TNF and RPKM must be Numpy arrays of dtype float32")
 
-    mask = tnf.sum(axis=1) != 0
+    ### Copy arrays and mask them ###
+    # Copy if not destroy - this way we can have all following operations in-place
+    # for simplicity
+    if not destroy:
+        rpkm = rpkm.copy()
+        tnf = tnf.copy()
+
+    # Normalize samples to have same depth
+    sample_depths_sum = rpkm.sum(axis=0)
+    if _np.any(sample_depths_sum == 0):
+        raise ValueError(
+            "One or more samples have zero depth in all sequences, so cannot be depth normalized"
+        )
+    rpkm *= 1_000_000 / sample_depths_sum
 
     # If multiple samples, also include nonzero depth as requirement for accept
     # of sequences
+    mask = tnf.sum(axis=1) != 0
+    depthssum = None
     if rpkm.shape[1] > 1:
         depthssum = rpkm.sum(axis=1)
         mask &= depthssum != 0
         depthssum = depthssum[mask]
+        assert isinstance(depthssum, _np.ndarray)
 
     if mask.sum() < batchsize:
-        raise ValueError('Fewer sequences left after filtering than the batch size.')
+        raise ValueError(
+            "Fewer sequences left after filtering than the batch size. "
+            + "This probably means you try to run on a too small dataset (below ~10k sequences), "
+            + "or that nearly all sequences were filtered away. Check the log file, "
+            + "and verify BAM file content is sensible."
+        )
 
-    if destroy:
-        rpkm = _vambtools.numpy_inplace_maskarray(rpkm, mask)
-        tnf = _vambtools.numpy_inplace_maskarray(tnf, mask)
-    else:
-        # The astype operation does not copy due to "copy=False", but the masking
-        # operation does.
-        rpkm = rpkm[mask].astype(_np.float32, copy=False)
-        tnf = tnf[mask].astype(_np.float32, copy=False)
+    _vambtools.numpy_inplace_maskarray(rpkm, mask)
+    _vambtools.numpy_inplace_maskarray(tnf, mask)
 
     # If multiple samples, normalize to sum to 1, else zscore normalize
     if rpkm.shape[1] > 1:
+        assert depthssum is not None  # we set it so just above
         rpkm /= depthssum.reshape((-1, 1))
     else:
         _vambtools.zscore(rpkm, axis=0, inplace=True)
 
-    # Normalize arrays and create the Tensors (the tensors share the underlying memory)
-    # of the Numpy arrays
+    # Normalize TNF
     _vambtools.zscore(tnf, axis=0, inplace=True)
-    depthstensor = _torch.from_numpy(rpkm)
-    tnftensor = _torch.from_numpy(tnf)
 
-    # Create dataloader
+    # Create weights
+    lengths = (lengths[mask]).astype(_np.float32)
+    weights = _np.log(lengths).astype(_np.float32) - 5.0
+    weights[weights < 2.0] = 2.0
+    weights *= len(weights) / weights.sum()
+    weights.shape = (len(weights), 1)
+
+    ### Create final tensors and dataloader ###
+    depthstensor = _torch.from_numpy(rpkm)  # this is a no-copy operation
+    tnftensor = _torch.from_numpy(tnf)
+    weightstensor = _torch.from_numpy(weights)
     n_workers = 4 if cuda else 1
-    dataset = _TensorDataset(depthstensor, tnftensor)
-    dataloader = _DataLoader(dataset=dataset, batch_size=batchsize, drop_last=True,
-                             shuffle=True, num_workers=n_workers, pin_memory=cuda)
+    dataset = _TensorDataset(depthstensor, tnftensor, weightstensor)
+    dataloader = _DataLoader(
+        dataset=dataset,
+        batch_size=batchsize,
+        drop_last=True,
+        shuffle=True,
+        num_workers=n_workers,
+        pin_memory=cuda,
+    )
 
     return dataloader, mask
 
+
 class VAE(_nn.Module):
     """Variational autoencoder, subclass of torch.nn.Module.
 
     Instantiate with:
         nsamples: Number of samples in abundance matrix
-        nhiddens: List of n_neurons in the hidden layers [None=Auto]
+        nhiddens: list of n_neurons in the hidden layers [None=Auto]
         nlatent: Number of neurons in the latent layer [32]
         alpha: Approximate starting TNF/(CE+TNF) ratio in loss. [None = Auto]
         beta: Multiply KLD by the inverse of this value [200]
@@ -125,13 +178,21 @@ class VAE(_nn.Module):
     0.99 and 0.0, respectively
     """
 
-    def __init__(self, nsamples, nhiddens=None, nlatent=32, alpha=None,
-                 beta=200, dropout=0.2, cuda=False):
+    def __init__(
+        self,
+        nsamples: int,
+        nhiddens: Optional[list[int]] = None,
+        nlatent: int = 32,
+        alpha: Optional[float] = None,
+        beta: float = 200.0,
+        dropout: Optional[float] = 0.2,
+        cuda: bool = False,
+    ):
         if nlatent < 1:
-            raise ValueError('Minimum 1 latent neuron, not {}'.format(latent))
+            raise ValueError(f"Minimum 1 latent neuron, not {nlatent}")
 
         if nsamples < 1:
-            raise ValueError('nsamples must be > 0, not {}'.format(nsamples))
+            raise ValueError(f"nsamples must be > 0, not {nsamples}")
 
         # If only 1 sample, we weigh alpha and nhiddens differently
         if alpha is None:
@@ -144,16 +205,16 @@ class VAE(_nn.Module):
             dropout = 0.2 if nsamples > 1 else 0.0
 
         if any(i < 1 for i in nhiddens):
-            raise ValueError('Minimum 1 neuron per layer, not {}'.format(min(nhiddens)))
+            raise ValueError(f"Minimum 1 neuron per layer, not {min(nhiddens)}")
 
         if beta <= 0:
-            raise ValueError('beta must be > 0, not {}'.format(beta))
+            raise ValueError(f"beta must be > 0, not {beta}")
 
         if not (0 < alpha < 1):
-            raise ValueError('alpha must be 0 < alpha < 1, not {}'.format(alpha))
+            raise ValueError(f"alpha must be 0 < alpha < 1, not {alpha}")
 
         if not (0 <= dropout < 1):
-            raise ValueError('dropout must be 0 <= dropout < 1, not {}'.format(dropout))
+            raise ValueError(f"dropout must be 0 <= dropout < 1, not {dropout}")
 
         super(VAE, self).__init__()
 
@@ -174,7 +235,9 @@ class VAE(_nn.Module):
         self.decodernorms = _nn.ModuleList()
 
         # Add all other hidden layers
-        for nin, nout in zip([self.nsamples + self.ntnf] + self.nhiddens, self.nhiddens):
+        for nin, nout in zip(
+            [self.nsamples + self.ntnf] + self.nhiddens, self.nhiddens
+        ):
             self.encoderlayers.append(_nn.Linear(nin, nout))
             self.encodernorms.append(_nn.BatchNorm1d(nout))
 
@@ -198,7 +261,7 @@ class VAE(_nn.Module):
         if cuda:
             self.cuda()
 
-    def _encode(self, tensor):
+    def _encode(self, tensor: Tensor) -> tuple[Tensor, Tensor]:
         tensors = list()
 
         # Hidden layers
@@ -223,7 +286,7 @@ class VAE(_nn.Module):
         return mu, logsigma
 
     # sample with gaussian noise
-    def reparameterize(self, mu, logsigma):
+    def reparameterize(self, mu: Tensor, logsigma: Tensor) -> Tensor:
         epsilon = _torch.randn(mu.size(0), mu.size(1))
 
         if self.usecuda:
@@ -232,11 +295,11 @@ class VAE(_nn.Module):
         epsilon.requires_grad = True
 
         # See comment above regarding softplus
-        latent = mu + epsilon * _torch.exp(logsigma/2)
+        latent = mu + epsilon * _torch.exp(logsigma / 2)
 
         return latent
 
-    def _decode(self, tensor):
+    def _decode(self, tensor: Tensor) -> tuple[Tensor, Tensor]:
         tensors = list()
 
         for decoderlayer, decodernorm in zip(self.decoderlayers, self.decodernorms):
@@ -255,7 +318,9 @@ class VAE(_nn.Module):
 
         return depths_out, tnf_out
 
-    def forward(self, depths, tnf):
+    def forward(
+        self, depths: Tensor, tnf: Tensor
+    ) -> tuple[Tensor, Tensor, Tensor, Tensor]:
         tensor = _torch.cat((depths, tnf), 1)
         mu, logsigma = self._encode(tensor)
         latent = self.reparameterize(mu, logsigma)
@@ -263,54 +328,69 @@ class VAE(_nn.Module):
 
         return depths_out, tnf_out, mu, logsigma
 
-    def calc_loss(self, depths_in, depths_out, tnf_in, tnf_out, mu, logsigma):
+    def calc_loss(
+        self,
+        depths_in: Tensor,
+        depths_out: Tensor,
+        tnf_in: Tensor,
+        tnf_out: Tensor,
+        mu: Tensor,
+        logsigma: Tensor,
+        weights: Tensor,
+    ) -> tuple[Tensor, Tensor, Tensor, Tensor]:
         # If multiple samples, use cross entropy, else use SSE for abundance
         if self.nsamples > 1:
             # Add 1e-9 to depths_out to avoid numerical instability.
-            ce = - ((depths_out + 1e-9).log() * depths_in).sum(dim=1).mean()
+            ce = -((depths_out + 1e-9).log() * depths_in).sum(dim=1)
             ce_weight = (1 - self.alpha) / _log(self.nsamples)
         else:
-            ce = (depths_out - depths_in).pow(2).sum(dim=1).mean()
+            ce = (depths_out - depths_in).pow(2).sum(dim=1)
             ce_weight = 1 - self.alpha
 
-        sse = (tnf_out - tnf_in).pow(2).sum(dim=1).mean()
-        kld = -0.5 * (1 + logsigma - mu.pow(2) - logsigma.exp()).sum(dim=1).mean()
+        sse = (tnf_out - tnf_in).pow(2).sum(dim=1)
+        kld = -0.5 * (1 + logsigma - mu.pow(2) - logsigma.exp()).sum(dim=1)
         sse_weight = self.alpha / self.ntnf
         kld_weight = 1 / (self.nlatent * self.beta)
-        loss = ce * ce_weight + sse * sse_weight + kld * kld_weight
+        reconstruction_loss = ce * ce_weight + sse * sse_weight
+        kld_loss = kld * kld_weight
+        loss = (reconstruction_loss + kld_loss) * weights
 
-        return loss, ce, sse, kld
+        return loss.mean(), ce.mean(), sse.mean(), kld.mean()
 
-    def trainepoch(self, data_loader, epoch, optimizer, batchsteps, logfile):
+    def trainepoch(
+        self,
+        data_loader: _DataLoader,
+        epoch: int,
+        optimizer,
+        batchsteps: list[int],
+        logfile,
+    ) -> _DataLoader[tuple[Tensor, Tensor, Tensor]]:
         self.train()
 
-        epoch_loss = 0
-        epoch_kldloss = 0
-        epoch_sseloss = 0
-        epoch_celoss = 0
+        epoch_loss = 0.0
+        epoch_kldloss = 0.0
+        epoch_sseloss = 0.0
+        epoch_celoss = 0.0
 
         if epoch in batchsteps:
-            data_loader = _DataLoader(dataset=data_loader.dataset,
-                                      batch_size=data_loader.batch_size * 2,
-                                      shuffle=True,
-                                      drop_last=True,
-                                      num_workers=data_loader.num_workers,
-                                      pin_memory=data_loader.pin_memory)
+            data_loader =  set_batchsize(data_loader, data_loader.batch_size * 2)
 
-        for depths_in, tnf_in in data_loader:
+        for depths_in, tnf_in, weights in data_loader:
             depths_in.requires_grad = True
             tnf_in.requires_grad = True
 
             if self.usecuda:
                 depths_in = depths_in.cuda()
                 tnf_in = tnf_in.cuda()
+                weights = weights.cuda()
 
             optimizer.zero_grad()
 
             depths_out, tnf_out, mu, logsigma = self(depths_in, tnf_in)
 
-            loss, ce, sse, kld = self.calc_loss(depths_in, depths_out, tnf_in,
-                                                  tnf_out, mu, logsigma)
+            loss, ce, sse, kld = self.calc_loss(
+                depths_in, depths_out, tnf_in, tnf_out, mu, logsigma, weights
+            )
 
             loss.backward()
             optimizer.step()
@@ -321,20 +401,24 @@ class VAE(_nn.Module):
             epoch_celoss += ce.data.item()
 
         if logfile is not None:
-            print('\tEpoch: {}\tLoss: {:.6f}\tCE: {:.7f}\tSSE: {:.6f}\tKLD: {:.4f}\tBatchsize: {}'.format(
-                  epoch + 1,
-                  epoch_loss / len(data_loader),
-                  epoch_celoss / len(data_loader),
-                  epoch_sseloss / len(data_loader),
-                  epoch_kldloss / len(data_loader),
-                  data_loader.batch_size,
-                  ), file=logfile)
+            print(
+                "\tEpoch: {}\tLoss: {:.6f}\tCE: {:.7f}\tSSE: {:.6f}\tKLD: {:.4f}\tBatchsize: {}".format(
+                    epoch + 1,
+                    epoch_loss / len(data_loader),
+                    epoch_celoss / len(data_loader),
+                    epoch_sseloss / len(data_loader),
+                    epoch_kldloss / len(data_loader),
+                    data_loader.batch_size,
+                ),
+                file=logfile,
+            )
 
             logfile.flush()
 
+        self.eval()
         return data_loader
 
-    def encode(self, data_loader):
+    def encode(self, data_loader) -> _np.ndarray:
         """Encode a data loader to a latent representation with VAE
 
         Input: data_loader: As generated by train_vae
@@ -344,14 +428,9 @@ class VAE(_nn.Module):
 
         self.eval()
 
-        new_data_loader = _DataLoader(dataset=data_loader.dataset,
-                                      batch_size=data_loader.batch_size,
-                                      shuffle=False,
-                                      drop_last=False,
-                                      num_workers=1,
-                                      pin_memory=data_loader.pin_memory)
+        new_data_loader = set_batchsize(data_loader, data_loader.batch_size, encode=True)
 
-        depths_array, tnf_array = data_loader.dataset.tensors
+        depths_array, _, _ = data_loader.dataset.tensors
         length = len(depths_array)
 
         # We make a Numpy array instead of a Torch array because, if we create
@@ -361,19 +440,19 @@ class VAE(_nn.Module):
 
         row = 0
         with _torch.no_grad():
-            for depths, tnf in new_data_loader:
+            for depths, tnf, weights in new_data_loader:
                 # Move input to GPU if requested
                 if self.usecuda:
                     depths = depths.cuda()
                     tnf = tnf.cuda()
 
                 # Evaluate
-                out_depths, out_tnf, mu, logsigma = self(depths, tnf)
+                _, _, mu, _ = self(depths, tnf)
 
                 if self.usecuda:
                     mu = mu.cpu()
 
-                latent[row: row + len(mu)] = mu
+                latent[row : row + len(mu)] = mu
                 row += len(mu)
 
         assert row == length
@@ -385,19 +464,22 @@ class VAE(_nn.Module):
         Input: Path or binary opened filehandle
         Output: None
         """
-        state = {'nsamples': self.nsamples,
-                 'alpha': self.alpha,
-                 'beta': self.beta,
-                 'dropout': self.dropout,
-                 'nhiddens': self.nhiddens,
-                 'nlatent': self.nlatent,
-                 'state': self.state_dict(),
-                }
+        state = {
+            "nsamples": self.nsamples,
+            "alpha": self.alpha,
+            "beta": self.beta,
+            "dropout": self.dropout,
+            "nhiddens": self.nhiddens,
+            "nlatent": self.nlatent,
+            "state": self.state_dict(),
+        }
 
         _torch.save(state, filehandle)
 
     @classmethod
-    def load(cls, path, cuda=False, evaluate=True):
+    def load(
+        cls, path: Union[IO[bytes], str], cuda: bool = False, evaluate: bool = True
+    ):
         """Instantiates a VAE from a model file.
 
         Inputs:
@@ -412,13 +494,13 @@ class VAE(_nn.Module):
         # Forcably load to CPU even if model was saves as GPU model
         dictionary = _torch.load(path, map_location=lambda storage, loc: storage)
 
-        nsamples = dictionary['nsamples']
-        alpha = dictionary['alpha']
-        beta = dictionary['beta']
-        dropout = dictionary['dropout']
-        nhiddens = dictionary['nhiddens']
-        nlatent = dictionary['nlatent']
-        state = dictionary['state']
+        nsamples = dictionary["nsamples"]
+        alpha = dictionary["alpha"]
+        beta = dictionary["beta"]
+        dropout = dictionary["dropout"]
+        nhiddens = dictionary["nhiddens"]
+        nlatent = dictionary["nlatent"]
+        state = dictionary["state"]
 
         vae = cls(nsamples, nhiddens, nlatent, alpha, beta, dropout, cuda)
         vae.load_state_dict(state)
@@ -431,8 +513,15 @@ class VAE(_nn.Module):
 
         return vae
 
-    def trainmodel(self, dataloader, nepochs=500, lrate=1e-3,
-                   batchsteps=[25, 75, 150, 300], logfile=None, modelfile=None):
+    def trainmodel(
+        self,
+        dataloader: _DataLoader[tuple[Tensor, Tensor, Tensor]],
+        nepochs: int = 500,
+        lrate: float = 1e-3,
+        batchsteps: Optional[list[int]] = [25, 75, 150, 300],
+        logfile: Optional[IO[str]] = None,
+        modelfile: Union[None, str, Path, IO[bytes]] = None,
+    ):
         """Train the autoencoder from depths array and tnf array.
 
         Inputs:
@@ -447,50 +536,65 @@ class VAE(_nn.Module):
         """
 
         if lrate < 0:
-            raise ValueError('Learning rate must be positive, not {}'.format(lrate))
+            raise ValueError(f"Learning rate must be positive, not {lrate}")
 
         if nepochs < 1:
-            raise ValueError('Minimum 1 epoch, not {}'.format(nepochs))
+            raise ValueError("Minimum 1 epoch, not {nepochs}")
 
         if batchsteps is None:
-            batchsteps_set = set()
+            batchsteps_set: set[int] = set()
         else:
             # First collect to list in order to allow all element types, then check that
             # they are integers
             batchsteps = list(batchsteps)
             if not all(isinstance(i, int) for i in batchsteps):
-                raise ValueError('All elements of batchsteps must be integers')
+                raise ValueError("All elements of batchsteps must be integers")
             if max(batchsteps, default=0) >= nepochs:
-                raise ValueError('Max batchsteps must not equal or exceed nepochs')
-            last_batchsize = dataloader.batch_size * 2**len(batchsteps)
-            if len(dataloader.dataset) < last_batchsize:
-                raise ValueError('Last batch size exceeds dataset length')
+                raise ValueError("Max batchsteps must not equal or exceed nepochs")
+            last_batchsize = dataloader.batch_size * 2 ** len(batchsteps)
+            if len(dataloader.dataset) < last_batchsize:  # type: ignore
+                raise ValueError(
+                    f"Last batch size of {last_batchsize} exceeds dataset length "
+                    f"of {len(dataloader.dataset)}. "  # type: ignore
+                    "This means you have too few contigs left after filtering to train. "
+                    "It is not adviced to run Vamb with fewer than 10,000 sequences "
+                    "after filtering. "
+                    "Please check the Vamb log file to see where the sequences were "
+                    "filtered away, and verify BAM files has sensible content."
+                )
             batchsteps_set = set(batchsteps)
 
         # Get number of features
-        ncontigs, nsamples = dataloader.dataset.tensors[0].shape
+        # Following line is un-inferrable due to typing problems with DataLoader
+        ncontigs, nsamples = dataloader.dataset.tensors[0].shape  # type: ignore
         optimizer = _Adam(self.parameters(), lr=lrate)
 
         if logfile is not None:
-            print('\tNetwork properties:', file=logfile)
-            print('\tCUDA:', self.usecuda, file=logfile)
-            print('\tAlpha:', self.alpha, file=logfile)
-            print('\tBeta:', self.beta, file=logfile)
-            print('\tDropout:', self.dropout, file=logfile)
-            print('\tN hidden:', ', '.join(map(str, self.nhiddens)), file=logfile)
-            print('\tN latent:', self.nlatent, file=logfile)
-            print('\n\tTraining properties:', file=logfile)
-            print('\tN epochs:', nepochs, file=logfile)
-            print('\tStarting batch size:', dataloader.batch_size, file=logfile)
-            batchsteps_string = ', '.join(map(str, sorted(batchsteps))) if batchsteps_set else "None"
-            print('\tBatchsteps:', batchsteps_string, file=logfile)
-            print('\tLearning rate:', lrate, file=logfile)
-            print('\tN sequences:', ncontigs, file=logfile)
-            print('\tN samples:', nsamples, file=logfile, end='\n\n')
+            print("\tNetwork properties:", file=logfile)
+            print("\tCUDA:", self.usecuda, file=logfile)
+            print("\tAlpha:", self.alpha, file=logfile)
+            print("\tBeta:", self.beta, file=logfile)
+            print("\tDropout:", self.dropout, file=logfile)
+            print("\tN hidden:", ", ".join(map(str, self.nhiddens)), file=logfile)
+            print("\tN latent:", self.nlatent, file=logfile)
+            print("\n\tTraining properties:", file=logfile)
+            print("\tN epochs:", nepochs, file=logfile)
+            print("\tStarting batch size:", dataloader.batch_size, file=logfile)
+            batchsteps_string = (
+                ", ".join(map(str, sorted(batchsteps_set)))
+                if batchsteps_set
+                else "None"
+            )
+            print("\tBatchsteps:", batchsteps_string, file=logfile)
+            print("\tLearning rate:", lrate, file=logfile)
+            print("\tN sequences:", ncontigs, file=logfile)
+            print("\tN samples:", nsamples, file=logfile, end="\n\n")
 
         # Train
         for epoch in range(nepochs):
-            dataloader = self.trainepoch(dataloader, epoch, optimizer, batchsteps_set, logfile)
+            dataloader = self.trainepoch(
+                dataloader, epoch, optimizer, sorted(batchsteps_set), logfile
+            )
 
         # Save weights - Lord forgive me, for I have sinned when catching all exceptions
         if modelfile is not None:
